from ryu.base import app_manager
from ryu.controller import ofp_event
from ryu.controller.handler import CONFIG_DISPATCHER, MAIN_DISPATCHER
from ryu.controller.handler import set_ev_cls
from ryu.ofproto import ofproto_v1_3
from ryu.lib.packet import packet
from ryu.lib.packet import ethernet
import re



class SimpleSwitch13(app_manager.RyuApp):
	"""docstring for prova"""

	OFP_VERSIONS = [ofproto_v1_3.OFP_VERSION]

	def __init__(self, *args, **kwargs):
		super(SimpleSwitch13, self).__init__(*args, **kwargs)
		self.ipv6_to_port = {}



	@set_ev_cls(ofp_event.EventOFPSwitchFeatures, CONFIG_DISPATCHER)
	def switch_features_handler(self, ev):
		datapath = ev.msg.datapath
		ofproto = datapath.ofproto
		parser = datapath.ofproto_parser
		
		# install table-miss flow entry
		#
		# We specify NO BUFFER to max_len of the output action due to
		# OVS bug. At this moment, if we specify a lesser number, e.g.,
		# 128, OVS will send Packet-In with invalid buffer_id and
		# truncated packet data. In that case, we cannot output packets
		# correctly.
		match = parser.OFPMatch()
		actions = [parser.OFPActionOutput(ofproto.OFPP_CONTROLLER, ofproto.OFPCML_NO_BUFFER)]
		
		self.add_flow(datapath, 0, match, actions)


	@set_ev_cls(ofp_event.EventOFPPacketIn, MAIN_DISPATCHER)
	def _packet_in_handler(self, ev):
		msg = ev.msg
		datapath = msg.datapath
		ofproto = datapath.ofproto
		parser = datapath.ofproto_parser

		# id del datapath che ha mandato la PacketIn
		dpid = datapath.id

		# prendo l'indirizzo IPv6 del mittente e del destinatario
		srcAdd = msg.match['ipv6_src']

		dstAdd = msg.match['ipv6_dst']


		self.logger.info("packet in %s %s %s", dpid, srcAdd, dstAdd)



		# creo due regex per i due prefissi possibili
		p1 = re.compile('2001:1')
		p2 = re.compile('2002:2')

		m1 = p1.match(srcAdd)
		m2 = p2.match(srcAdd)


		if m1 != None:
			out_port = 2
		elif m2 != None:
			out_port = 3
		else:
			out_port = 0



		# creo un matching sul campo che mi interessa (?)
		match = parser.OFPMatch(ipv6_src=srcAdd)

		# azione da eseguire (flw mod)
		actions = [parser.OFPActionOutput(out_port)]

		# aggiungo la nuova regola al datapath
		self.add_flow(dpid, 1, match, actions)



		data = None
		if msg.buffer_id == ofproto.OFP_NO_BUFFER:
			data = msg.data

		# altra azione da eseguire (packet out)
		out = parser.OFPPacketOut(datapath=dpid, buffer_id=msg.buffer_id, actions=actions, data=data)

		# invia la packet out al datapath
		datapath.send_msg(out)



		
	def add_flow(self, datapath, priority, match, actions):
		ofproto = datapath.ofproto
		parser = datapath.ofproto_parser

		inst = [parser.OFPInstructionActions(ofproto.OFPIT_APPLY_ACTIONS, actions)]

		# crea il messaggio di flow modification
		mod = parser.OFPFlowMod(datapath=datapath, priority=priority, match=match, instructions=inst)
		
		# invia la nuova regola al datapath
		datapath.send_msg(mod)